from __future__ import annotations

import argparse
import pathlib
import sys

from importlib_metadata import version

from .core import run

ITERATIONS_DEFAULT = 50
EXTRA_DEFAULT = 1
COLOR_DEFAULT = 1.0
OUTPUT_DEFAULT = None

PROGRAM = "pycrusher"


def get_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog=PROGRAM)
    parser.add_argument(
        "input_path",
        type=pathlib.Path,
        help="Input image filename",
        metavar="INPUT_PATH",
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {version(PROGRAM)}",
        help="show %(prog)s version",
    )

    parser.add_argument(
        "-i",
        "--iterations",
        dest="iterations",
        type=int,
        help="Number of compression iterations",
        default=ITERATIONS_DEFAULT,
    )
    parser.add_argument(
        "-e",
        "--extra",
        dest="extra",
        type=int,
        help="Number of nested iterations",
        default=EXTRA_DEFAULT,
    )
    parser.add_argument(
        "-r",
        "--reverse",
        dest="reverse",
        action="store_true",
        help="Reverses compression iterations (lower to higher)",
    )

    parser.add_argument(
        "-c",
        "--color",
        dest="color",
        type=float,
        help="Color enhancement",
        default=COLOR_DEFAULT,
    )

    parser.add_argument(
        "-o",
        "--output",
        dest="output_path",
        type=pathlib.Path,
        help="Output image filename",
        default=OUTPUT_DEFAULT,
    )

    parser.add_argument(
        "-p",
        "--preprocess",
        dest="preprocess",
        action="store_true",
        help="Adds color enhancement BEFORE compression",
    )

    return parser


def validate_input_path(namespace: argparse.Namespace) -> None:
    if not namespace.input_path.exists():
        msg = f"Input path does not exist: {namespace.input_path}"
        raise TypeError(msg)

    if not namespace.input_path.is_file():
        msg = "Input path should be a file."
        raise TypeError(msg)


def validate_iterations(namespace: argparse.Namespace) -> None:
    if namespace.iterations <= 0:
        msg = f"Iterations must be greater or equal to 1: {namespace.iterations}"
        raise TypeError(msg)


def validate_extra(namespace: argparse.Namespace) -> None:
    if namespace.extra <= 0:
        msg = f"Extra must be greater or equal to 1: {namespace.extra}"
        raise TypeError(msg)


def validate_color(namespace: argparse.Namespace) -> None:
    if namespace.color < 0.0:
        msg = f"Color enhancement must be greater or equal to 0.0: {namespace.color}"
        raise TypeError(msg)


def main() -> None:
    parser = get_argparser()

    namespace = parser.parse_args(sys.argv[1:])

    validate_input_path(namespace)
    validate_iterations(namespace)
    validate_extra(namespace)
    validate_color(namespace)

    run(
        input_path=namespace.input_path,
        output_path=namespace.output_path,
        iterations=namespace.iterations,
        extra=namespace.extra,
        color=namespace.color,
        reverse=namespace.reverse,
        preprocess=namespace.preprocess,
    )

    print("Done!")  # noqa: T201



from __future__ import annotations

import io
import pathlib
from typing import TYPE_CHECKING, Generator, TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterable

import hypothesis
import pytest
from hypothesis import strategies as st

from pycrusher.core import compress, generate_quality_sequence
from tests.utils import SMALL_TEST_IMAGES_DIRECTORY, same_pixels_in_image_files


class TestGetQualities:
    T = TypeVar("T")

    # itertools.pairwise was added in python 3.10
    @staticmethod
    def pairwise(iterable: Iterable[T]) -> Generator[tuple[T | None, T], T, None]:
        iterator = iter(iterable)
        a = next(iterator, None)
        for b in iterator:
            yield a, b
            a = b

    @hypothesis.given(
        iterations=st.integers(min_value=1, max_value=10),
        reverse=st.booleans(),
    )
    def test_get_qualities(
        self,
        iterations: int,
        reverse: bool,
    ) -> None:
        qualities = generate_quality_sequence(iterations, reverse=reverse)
        assert len(qualities) == iterations

        if len(qualities) >= 1:
            if reverse:
                assert qualities[0] == 0
            else:
                assert qualities[0] == 100

        if len(qualities) > 1:
            delta = 100 // len(qualities)
            for a, b in self.pairwise(qualities):
                if reverse:
                    assert a < b
                else:
                    assert a > b
                assert abs(a - b) == delta


class TestGetDefaultOutputName:
    def test_get_default_output_name(self) -> None:
        pass


class TestChangeColors:
    def test_change_color(self) -> None:
        pass


class TestCompress:
    @hypothesis.settings(deadline=None)
    @pytest.mark.parametrize("input_path", SMALL_TEST_IMAGES_DIRECTORY.iterdir())
    @hypothesis.given(
        iterations=st.integers(min_value=1, max_value=2),
        extra=st.integers(min_value=1, max_value=2),
        reverse=st.booleans(),
    )
    def test_compress_raises_no_exceptions(
        self,
        input_path: pathlib.Path,
        iterations: int,
        extra: int,
        reverse: bool,
    ) -> None:
        buf = io.BytesIO(input_path.read_bytes())

        qualities = extra * generate_quality_sequence(iterations, reverse)

        compress(
            buf,
            qualities,
        )

    @hypothesis.settings(deadline=None)
    @pytest.mark.parametrize("input_path", SMALL_TEST_IMAGES_DIRECTORY.iterdir())
    @hypothesis.given(
        iterations=st.integers(min_value=2, max_value=5),
        extra=st.integers(min_value=2, max_value=3),
        reverse=st.booleans(),
    )
    def test_compress(
        self,
        input_path: pathlib.Path,
        iterations: int,
        extra: int,
        reverse: bool,
    ) -> None:
        buf = io.BytesIO(input_path.read_bytes())

        qualities = extra * generate_quality_sequence(iterations, reverse)

        compress(
            buf,
            qualities,
        )

        assert not same_pixels_in_image_files(input_path, buf)


def test_run():
    pass



from __future__ import annotations

import argparse
import pathlib

import hypothesis
import pytest
from hypothesis import strategies as st
from importlib_metadata import version

from pycrusher.cli import (
    COLOR_DEFAULT,
    EXTRA_DEFAULT,
    ITERATIONS_DEFAULT,
    OUTPUT_DEFAULT,
    PROGRAM,
    get_argparser,
    validate_color,
    validate_extra,
    validate_input_path,
    validate_iterations,
)
from tests.utils import SMALL_TEST_IMAGES_DIRECTORY


class TestParseArgs:
    def test_input_path(self, capsys: pytest.CaptureFixture[str]) -> None:
        # Input_path is required
        # Parsed as path
        parser = get_argparser()
        namespace = parser.parse_args(["placeholder_path"])
        assert namespace.input_path == pathlib.Path("placeholder_path")
        with pytest.raises(SystemExit):
            # Input_path is the only required argument
            parser.parse_args([])

    @pytest.mark.parametrize("arg", ["-i", "--iterations"])
    @hypothesis.given(iterations=st.integers())
    def test_iterations(self, arg: str, iterations: int) -> None:
        # Extra is optional
        # Extra is integer
        # -e, or --extra
        # If missing, default is 1
        parser = get_argparser()

        namespace = parser.parse_args(["placeholder_path", arg, str(iterations)])
        assert namespace.iterations == iterations

        namespace = parser.parse_args(["placeholder_path"])
        assert namespace.iterations == ITERATIONS_DEFAULT

    @pytest.mark.parametrize("arg", ["-e", "--extra"])
    @hypothesis.given(extra=st.integers())
    def test_extra(self, arg: str, extra: int) -> None:
        # Extra is optional
        # Extra is integer
        # -e, or --extra
        # If missing, default is 1
        parser = get_argparser()

        namespace = parser.parse_args(["placeholder_path", arg, str(extra)])
        assert namespace.extra == extra

        namespace = parser.parse_args(["placeholder_path"])
        assert namespace.extra == EXTRA_DEFAULT

    @pytest.mark.parametrize("arg", ["-c", "--color"])
    @hypothesis.given(color=st.floats(min_value=0.0))
    def test_color(self, arg: str, color: float) -> None:
        # Color is optional
        # Color is float
        # -c, or --color
        # If missing, default is 1.0
        parser = get_argparser()

        namespace = parser.parse_args(["placeholder_path", arg, str(color)])
        assert namespace.color == color

        namespace = parser.parse_args(["placeholder_path"])
        assert namespace.color == COLOR_DEFAULT

    @pytest.mark.parametrize("arg", ["-o", "--output"])
    def test_output_path(self, arg: str) -> None:
        # Output is optional
        # Output is Path
        # -o or --output
        # If missing, default is None
        output_path = pathlib.Path("foo")
        parser = get_argparser()

        namespace = parser.parse_args(["placeholder_path", arg, str(output_path)])
        assert namespace.output_path == output_path

        namespace = parser.parse_args(["placeholder_path"])
        assert namespace.output_path == OUTPUT_DEFAULT

    @pytest.mark.parametrize("arg", ["-r", "--reverse"])
    def test_reverse(self, arg: str) -> None:
        # Reverse is optional
        # Reverse is bool
        # -r or --reverse
        # If missing, False. Else, true
        parser = get_argparser()

        namespace = parser.parse_args(["placeholder_path", arg])
        assert namespace.reverse

        namespace = parser.parse_args(["placeholder_path"])
        assert not namespace.reverse

    @pytest.mark.parametrize("arg", ["-p", "--preprocess"])
    def test_preprocess(self, arg: str) -> None:
        # Preprocess is optional
        # Preprocess is bool
        # -p or --preprocess
        # If missing, false. Else, true
        parser = get_argparser()

        namespace = parser.parse_args(["placeholder_path", arg])
        assert namespace.preprocess

        namespace = parser.parse_args(["placeholder_path"])
        assert not namespace.preprocess

    @hypothesis.given(st.permutations(["-i 1", "-e 1", "-c 1", "-r", "-p", "-o out"]))
    def test_input_path_in_edges_of_args(self, permutated_args: list[str]) -> None:
        parser = get_argparser()

        joined_args1 = " ".join(["placeholder_path", *permutated_args])
        joined_args2 = " ".join([*permutated_args, "placeholder_path"])

        namespace1 = parser.parse_args(joined_args1.split())
        namespace2 = parser.parse_args(joined_args2.split())

        expected_namespace = argparse.Namespace(
            input_path=pathlib.Path("placeholder_path"),
            iterations=1,
            extra=1,
            color=1.0,
            reverse=True,
            preprocess=True,
            output_path=pathlib.Path("out"),
        )
        assert namespace1 == expected_namespace
        assert namespace2 == expected_namespace

    def test_cli_help_exits_with_retcode_zero(
        self,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """
        Test that '-h' and '--help' behave the same:
        They raise SystemExit(0) and output the same message.
        """
        parser = get_argparser()

        with pytest.raises(SystemExit) as exc_info:
            parser.parse_args(["-h"])
        assert exc_info.value.args[0] == 0
        capture1 = capsys.readouterr()

        with pytest.raises(SystemExit) as exc_info:
            parser.parse_args(["--help"])
        assert exc_info.value.args[0] == 0
        capture2 = capsys.readouterr()

        assert capture1.out == capture2.out

    def test_cli_version_exits_with_retcode_zero(
        self,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """
        Test that '-v' and '--version' behave the same:
        Both raise SystemExit(0) and output the same message.
        """
        argparser = get_argparser()
        with pytest.raises(SystemExit) as exc_info:
            argparser.parse_args(["-v"])
        assert exc_info.value.args[0] == 0
        capture1 = capsys.readouterr()

        with pytest.raises(SystemExit) as exc_info:
            argparser.parse_args(["--version"])
        assert exc_info.value.args[0] == 0
        capture2 = capsys.readouterr()

        assert capture1.out == capture2.out

        assert capture1.out == f"{PROGRAM} {version(PROGRAM)}\n"


class TestValidateNamespace:
    @pytest.mark.parametrize("input_path", SMALL_TEST_IMAGES_DIRECTORY.iterdir())
    def test_valid_input_path(self, input_path: pathlib.Path) -> None:
        validate_input_path(argparse.Namespace(input_path=input_path))

    @hypothesis.given(st.uuids().map(lambda u: pathlib.Path(str(u))))
    def test_invalid_input_path_does_not_exist(self, input_path: pathlib.Path) -> None:
        with pytest.raises(
            TypeError,
            match="Input path does not exist:",
        ):
            validate_input_path(argparse.Namespace(input_path=input_path))

    def test_invalid_input_path_is_not_directory(self, tmpdir: pathlib.Path) -> None:
        with pytest.raises(
            TypeError,
            match="Input path should be a file.",
        ):
            validate_input_path(argparse.Namespace(input_path=pathlib.Path(tmpdir)))

    @hypothesis.given(iterations=st.integers(max_value=0))
    def test_invalid_iterations(self, iterations: int) -> None:
        with pytest.raises(
            TypeError,
            match="Iterations must be greater or equal to 1:",
        ):
            validate_iterations(argparse.Namespace(iterations=iterations))

    @hypothesis.given(iterations=st.integers(min_value=1))
    def test_valid_iterations(self, iterations: int) -> None:
        validate_iterations(argparse.Namespace(iterations=iterations))

    @hypothesis.given(extra=st.integers(max_value=0))
    def test_invalid_extra(self, extra: int) -> None:
        with pytest.raises(
            TypeError,
            match="Extra must be greater or equal to 1:",
        ):
            validate_extra(argparse.Namespace(extra=extra))

    @hypothesis.given(extra=st.integers(min_value=1))
    def test_valid_extra(self, extra: int) -> None:
        validate_extra(argparse.Namespace(extra=extra))

    @hypothesis.given(color=st.floats(max_value=-1e5))
    def test_invalid_color(self, color: float) -> None:
        with pytest.raises(
            TypeError,
            match="Color enhancement must be greater or equal to 0.0:",
        ):
            validate_color(argparse.Namespace(color=color))

    @hypothesis.given(color=st.floats(min_value=0))
    def test_valid_color(
        self,
        color: float,
    ) -> None:
        validate_color(argparse.Namespace(color=color))



from __future__ import annotations

import io
import pathlib

import tqdm
from PIL import Image, ImageEnhance

CWD_COMPRESSIONS_DIRECTORY = pathlib.Path.cwd().joinpath("compressions")


def generate_default_output_name(
    input_path: pathlib.Path,
    *,
    iterations: int,
    extra: int,
    color: float,
    reverse: bool,
    preprocess: bool,
) -> str:
    """
    Generate default output name based on pycrusher parameters.

    Args:
        input_path (pathlib.Path): Input given by parse_args.
        iterations (int): How many times to iterate compression.
        extra (int): How much to enforce compression.
        color (float): Color saturation.
        reverse (bool): Reverse qualities.
        preprocess (bool): Preprocess color.

    Returns:
        Default output name based on pycrusher parameters.

    """
    output_suffixes = [f"i{iterations}", f"e{extra}"]
    if reverse:
        output_suffixes.append("rev")
    if preprocess:
        output_suffixes.append("pre")
    if color != 1.0:
        output_suffixes.append(f"c{color}")

    joined_suffixes = "_".join(output_suffixes)
    return f"{input_path.stem}_{joined_suffixes}.jpg"


def compress(
    image_buffer: io.BytesIO,
    qualities: list[int],
) -> None:
    """
    Save file repeatedly as JPEG for each quality in qualities.

    Args:
        image_buffer (io.BytesIO): Buffer containing image file.
        qualities (list[int]): List of JPEG qualities.

    """
    for quality in tqdm.tqdm(qualities):
        with Image.open(image_buffer) as img:
            img.load()

            image_buffer.seek(0)
            image_buffer.truncate()

            img.save(
                image_buffer,
                format="JPEG",
                quality=quality,
            )


def change_color(
    image_buffer: io.BytesIO,
    color: float,
    quality: int,
) -> None:
    """
    Change image saturation and save it with last compression quality.

    Args:
        image_buffer (io.BytesIO): Buffer containing image file.
        color (float): Color enhancement factor
        quality (int): JPEG quality

    """
    with Image.open(image_buffer) as img:
        converter = ImageEnhance.Color(img)
        enhanced_img = converter.enhance(color)

        image_buffer.seek(0)
        image_buffer.truncate()

        enhanced_img.save(
            image_buffer,
            format="JPEG",
            quality=quality,
        )


def generate_quality_sequence(
    iterations: int,
    reverse: bool,
) -> list[int]:
    """
    Generate JPEG quality sequence.

    Args:
        iterations (int): Number of iterations
        reverse (bool): Reverse list?

    Returns:
        List of JPEG qualities

    """
    # Quality sequence changed a bit, such that
    # qualities have uniform spacing.
    delta = 100 // iterations
    if reverse:
        qualities = [delta * i for i in range(iterations)]
    else:
        qualities = [100 - (delta * i) for i in range(iterations)]
    return qualities


def confirm(title: str, question: str) -> bool:
    """
    Confirm action.

    User is required to respond.

    Args:
        title (str): Top message, printed only once.
        question (str): Question user must respond.

    Returns:
        Confirm (True) or deny (False) question.

    """
    print(title)  # noqa: T201
    while True:
        confirm = input(f"{question} (y/n) ")
        if not confirm:
            continue
        if confirm.lower() in {"n", "no"}:
            return False
        if confirm.lower() in {"y", "yes"}:
            return True


def run(
    *,
    input_path: pathlib.Path,
    iterations: int,
    extra: int,
    color: float,
    reverse: bool,
    preprocess: bool,
    output_path: pathlib.Path | None,
) -> None:
    if output_path is None:
        default_output_name = generate_default_output_name(
            input_path,
            iterations=iterations,
            extra=extra,
            color=color,
            reverse=reverse,
            preprocess=preprocess,
        )
        output_path = CWD_COMPRESSIONS_DIRECTORY.joinpath(default_output_name)
        CWD_COMPRESSIONS_DIRECTORY.mkdir(exist_ok=True)

    if output_path.exists():
        should_overwrite = confirm(
            title=f"File already exists: {output_path}",
            question="Do you want to overwrite it?",
        )
        if not should_overwrite:
            return

    qualities = extra * generate_quality_sequence(iterations, reverse)

    with io.BytesIO(input_path.read_bytes()) as image_buffer:
        if preprocess:
            change_color(image_buffer, color, quality=qualities[0])
            compress(image_buffer, qualities[1:])
        else:
            compress(image_buffer, qualities[:-1])
            change_color(image_buffer, color, quality=qualities[-1])

        output_path.write_bytes(image_buffer.getvalue())
